package rpc

import (
	"context"
	"errors"
	"testing"

	"github.com/rs/zerolog"
	"github.com/stretchr/testify/mock"
	"github.com/stretchr/testify/suite"
	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/status"

	"github.com/onflow/flow/protobuf/go/flow/entities"
	"github.com/onflow/flow/protobuf/go/flow/execution"

	"github.com/onflow/flow-go/engine/common/rpc/convert"
	realingestion "github.com/onflow/flow-go/engine/execution/ingestion"
	ingestion "github.com/onflow/flow-go/engine/execution/ingestion/mock"
	"github.com/onflow/flow-go/model/flow"
	realstorage "github.com/onflow/flow-go/storage"
	storage "github.com/onflow/flow-go/storage/mock"
	"github.com/onflow/flow-go/utils/unittest"
)

type Suite struct {
	suite.Suite
	log        zerolog.Logger
	events     *storage.Events
	exeResults *storage.ExecutionResults
	txResults  *storage.TransactionResults
	blocks     *storage.Blocks
}

func TestHandler(t *testing.T) {
	suite.Run(t, new(Suite))
}

func (suite *Suite) SetupTest() {
	suite.log = zerolog.Logger{}
	suite.events = new(storage.Events)
	suite.exeResults = new(storage.ExecutionResults)
	suite.txResults = new(storage.TransactionResults)
	suite.blocks = new(storage.Blocks)
}

// TestGetEventsForBlockIDs tests the GetEventsForBlockIDs API call
func (suite *Suite) TestGetEventsForBlockIDs() {

	totalBlocks := 10
	eventsPerBlock := 10

	blockIDs := make([][]byte, totalBlocks)
	expectedResult := make([]*execution.GetEventsForBlockIDsResponse_Result, totalBlocks)

	// setup the events storage mock
	for i := range blockIDs {
		block := unittest.BlockFixture()
		block.Header.Height = uint64(i)
		id := block.ID()
		blockIDs[i] = id[:]
		eventsForBlock := make([]flow.Event, eventsPerBlock)
		eventMessages := make([]*entities.Event, eventsPerBlock)
		for j := range eventsForBlock {
			e := unittest.EventFixture(flow.EventAccountCreated, uint32(j), uint32(j), unittest.IdentifierFixture())
			eventsForBlock[j] = e
			eventMessages[j] = convert.EventToMessage(e)
		}
		// expect one call to lookup result for each block ID
		suite.exeResults.On("ByBlockID", id).Return(nil, nil).Once()

		// expect one call to lookup events for each block ID
		suite.events.On("ByBlockIDEventType", id, flow.EventAccountCreated).Return(eventsForBlock, nil).Once()

		// expect one call to lookup each block
		suite.blocks.On("ByID", id).Return(&block, nil).Once()

		// create the expected result for this block
		expectedResult[i] = &execution.GetEventsForBlockIDsResponse_Result{
			BlockId:     id[:],
			BlockHeight: block.Header.Height,
			Events:      eventMessages,
		}
	}

	// create the handler
	handler := &handler{
		blocks:             suite.blocks,
		events:             suite.events,
		exeResults:         suite.exeResults,
		transactionResults: suite.txResults,
		chain:              flow.Mainnet,
		scriptsEnabled:     true,
		queriesEnabled:     true,
	}

	createRequest := func(errType string, blockIDs [][]byte) *execution.GetEventsForBlockIDsRequest {
		return &execution.GetEventsForBlockIDsRequest{
			Type:     errType,
			BlockIds: blockIDs,
		}
	}

	// happy path - valid requests receives a valid response
	suite.Run("happy path", func() {

		// create a valid API request
		req := createRequest(string(flow.EventAccountCreated), blockIDs)

		// execute the GetEventsForBlockIDs call
		resp, err := handler.GetEventsForBlockIDs(context.Background(), req)

		// check that a successful response is received
		suite.Require().NoError(err)

		actualResult := resp.GetResults()
		suite.Require().ElementsMatch(expectedResult, actualResult)

		// check that appropriate storage calls were made
		suite.events.AssertExpectations(suite.T())
	})

	// failure path - empty even type in the request results in an error
	suite.Run("request with empty event type", func() {

		// create an API request with empty even type
		req := createRequest("", blockIDs)

		_, err := handler.GetEventsForBlockIDs(context.Background(), req)

		// check that an error was received
		suite.Require().Error(err)
		errors.Is(err, status.Error(codes.InvalidArgument, ""))

		// check that no storage calls was made
		suite.events.AssertExpectations(suite.T())
	})

	// failure path - empty block ids in request results in an error
	suite.Run("request with empty block IDs", func() {

		// create an API request with empty block ids
		req := createRequest(string(flow.EventAccountCreated), nil)

		_, err := handler.GetEventsForBlockIDs(context.Background(), req)

		// check that an error was received
		suite.Require().Error(err)
		errors.Is(err, status.Error(codes.InvalidArgument, ""))

		// check that no storage calls was made
		suite.events.AssertExpectations(suite.T())
	})

	// failure path - non-existent block id in request results in an error
	suite.Run("request with non-existent block ID", func() {

		id := unittest.IdentifierFixture()

		// expect a storage call for the invalid id but return an error
		suite.exeResults.On("ByBlockID", id).Return(nil, realstorage.ErrNotFound).Once()

		// create an API request with the invalid block id
		req := createRequest(string(flow.EventAccountCreated), [][]byte{id[:]})

		_, err := handler.GetEventsForBlockIDs(context.Background(), req)

		// check that an error was received
		suite.Require().Error(err)
		errors.Is(err, status.Error(codes.NotFound, ""))

		// check that no storage calls was made
		suite.events.AssertExpectations(suite.T())
	})
}

// Test GetAccountAtBlockID tests the GetAccountAtBlockID API call
func (suite *Suite) TestGetAccountAtBlockID() {

	id := unittest.IdentifierFixture()
	serviceAddress := flow.Mainnet.Chain().ServiceAddress()

	serviceAccount := flow.Account{
		Address: serviceAddress,
	}

	// create the handler
	createHandler := func(engine realingestion.IngestRPC, scriptsEnabled bool) *handler {
		handler := &handler{
			engine:         engine,
			chain:          flow.Mainnet,
			scriptsEnabled: scriptsEnabled,
			queriesEnabled: true,
		}
		return handler
	}

	createRequest := func(id []byte, address []byte) *execution.GetAccountAtBlockIDRequest {
		return &execution.GetAccountAtBlockIDRequest{
			Address: address,
			BlockId: id,
		}
	}

	suite.Run("happy path with valid request", func() {

		// setup mock expectations
		mockEngine := new(ingestion.IngestRPC)
		mockEngine.On("GetAccount", mock.Anything, serviceAddress, id).Return(&serviceAccount, nil).Once()

		handler := createHandler(mockEngine, true)

		req := createRequest(id[:], serviceAddress.Bytes())

		resp, err := handler.GetAccountAtBlockID(context.Background(), req)

		suite.Require().NoError(err)
		actualAccount := resp.GetAccount()
		expectedAccount, err := convert.AccountToMessage(&serviceAccount)
		suite.Require().NoError(err)
		suite.Require().Equal(*expectedAccount, *actualAccount)
		mockEngine.AssertExpectations(suite.T())
	})

	suite.Run("invalid request with nil block id", func() {

		mockEngine := new(ingestion.IngestRPC)
		handler := createHandler(mockEngine, true)

		req := createRequest(nil, serviceAddress.Bytes())

		_, err := handler.GetAccountAtBlockID(context.Background(), req)

		suite.Require().Error(err)
	})

	suite.Run("invalid request with nil root address", func() {

		mockEngine := new(ingestion.IngestRPC)
		handler := createHandler(mockEngine, true)

		req := createRequest(id[:], nil)

		_, err := handler.GetAccountAtBlockID(context.Background(), req)

		suite.Require().Error(err)
	})

	// failure path - method should fail if script executions are disabled
	suite.Run("request with script executions disabled", func() {

		req := createRequest(id[:], serviceAddress.Bytes())

		mockEngine := new(ingestion.IngestRPC)
		handler := createHandler(mockEngine, false)

		_, err := handler.GetAccountAtBlockID(context.Background(), req)

		// check that an UNIMPLEMENTED error was received
		suite.Require().Error(err)
		suite.Assert().Equal(codes.Unimplemented, status.Code(err))
	})
}

// TestGetTransactionResult tests the GetTransactionResult API call
func (suite *Suite) TestGetTransactionResult() {

	totalEvents := 10
	block := unittest.BlockFixture()
	tx := unittest.TransactionFixture()
	bID := block.ID()
	txID := tx.ID()

	// setup the events storage mock
	eventsForTx := make([]flow.Event, totalEvents)
	eventMessages := make([]*entities.Event, totalEvents)
	for j := range eventsForTx {
		e := unittest.EventFixture(flow.EventAccountCreated, uint32(j), uint32(j), unittest.IdentifierFixture())
		eventsForTx[j] = e
		eventMessages[j] = convert.EventToMessage(e)
	}

	// expect a call to lookup events by block ID and transaction ID
	suite.events.On("ByBlockIDTransactionID", bID, txID).Return(eventsForTx, nil)

	// expect a call to lookup each block
	suite.blocks.On("ByID", block.ID()).Return(&block, true)

	// create the handler
	createHandler := func(txResults *storage.TransactionResults, queriesEnabled bool) *handler {
		handler := &handler{
			blocks:             suite.blocks,
			events:             suite.events,
			transactionResults: txResults,
			chain:              flow.Mainnet,
			scriptsEnabled:     true,
			queriesEnabled:     queriesEnabled,
		}
		return handler
	}

	// createRequest creates a GetEventsForBlockIDTransactionIDRequest
	createRequest := func(blockID []byte, transactionID []byte) *execution.GetTransactionResultRequest {
		return &execution.GetTransactionResultRequest{
			BlockId:       blockID,
			TransactionId: transactionID,
		}
	}

	assertEqual := func(expected, actual *execution.GetTransactionResultResponse) {
		suite.Require().Equal(expected.GetStatusCode(), actual.GetStatusCode())
		suite.Require().Equal(expected.GetErrorMessage(), actual.GetErrorMessage())
		suite.Require().ElementsMatch(expected.GetEvents(), actual.GetEvents())
	}

	// happy path - valid requests receives all events for the given transaction
	suite.Run("happy path with valid events and no transaction error", func() {

		// create the expected result
		expectedResult := &execution.GetTransactionResultResponse{
			StatusCode:   0,
			ErrorMessage: "",
			Events:       eventMessages,
		}

		// expect a call to lookup transaction result by block ID and transaction ID, return a result with no error
		txResults := new(storage.TransactionResults)
		txResult := flow.TransactionResult{
			TransactionID: flow.Identifier{},
			ErrorMessage:  "",
		}
		txResults.On("ByBlockIDTransactionID", bID, txID).Return(&txResult, nil).Once()

		handler := createHandler(txResults, true)

		// create a valid API request
		req := createRequest(bID[:], txID[:])

		// execute the GetTransactionResult call
		actualResult, err := handler.GetTransactionResult(context.Background(), req)

		// check that a successful response is received
		suite.Require().NoError(err)

		// check that all fields in response are as expected
		assertEqual(expectedResult, actualResult)

		// check that appropriate storage calls were made
		suite.events.AssertExpectations(suite.T())
		txResults.AssertExpectations(suite.T())
	})

	// happy path - valid requests receives all events and an error for the given transaction
	suite.Run("happy path with valid events and a transaction error", func() {

		// create the expected result
		expectedResult := &execution.GetTransactionResultResponse{
			StatusCode:   1,
			ErrorMessage: "runtime error",
			Events:       eventMessages,
		}

		// setup the storage to return a transaction error
		txResults := new(storage.TransactionResults)
		txResult := flow.TransactionResult{
			TransactionID: txID,
			ErrorMessage:  "runtime error",
		}
		txResults.On("ByBlockIDTransactionID", bID, txID).Return(&txResult, nil).Once()

		handler := createHandler(txResults, true)

		// create a valid API request
		req := createRequest(bID[:], txID[:])

		// execute the GetEventsForBlockIDTransactionID call
		actualResult, err := handler.GetTransactionResult(context.Background(), req)

		// check that a successful response is received
		suite.Require().NoError(err)

		// check that all fields in response are as expected
		assertEqual(expectedResult, actualResult)

		// check that appropriate storage calls were made
		suite.events.AssertExpectations(suite.T())
		txResults.AssertExpectations(suite.T())
	})

	// failure path - nil transaction ID in the request results in an error
	suite.Run("request with nil tx ID", func() {

		// create an API request with transaction ID as nil
		req := createRequest(bID[:], nil)

		// expect a call to lookup transaction result by block ID and transaction ID, return an error
		txResults := new(storage.TransactionResults)

		txResults.On("ByBlockIDTransactionID", bID, nil).Return(nil, status.Error(codes.InvalidArgument, "")).Once()

		handler := createHandler(txResults, true)

		_, err := handler.GetTransactionResult(context.Background(), req)

		// check that an error was received
		suite.Require().Error(err)
		errors.Is(err, status.Error(codes.InvalidArgument, ""))

		// check that no storage calls was made
		suite.events.AssertExpectations(suite.T())
	})

	// failure path - nil block id in the request results in an error
	suite.Run("request with nil block ID", func() {

		// create an API request with a nil block id
		req := createRequest(nil, txID[:])

		txResults := new(storage.TransactionResults)

		txResults.On("ByBlockIDTransactionID", nil, txID).Return(nil, status.Error(codes.InvalidArgument, "")).Once()

		handler := createHandler(txResults, true)

		_, err := handler.GetTransactionResult(context.Background(), req)

		// check that an error was received
		suite.Require().Error(err)
		errors.Is(err, status.Error(codes.InvalidArgument, ""))

		// check that no storage calls was made
		suite.events.AssertExpectations(suite.T())
	})

	// failure path - non-existent transaction ID in request results in an error
	suite.Run("request with non-existent transaction ID", func() {

		wrongTxID := unittest.IdentifierFixture()

		// create an API request with the invalid transaction ID
		req := createRequest(bID[:], wrongTxID[:])

		// expect a storage call for the invalid tx ID but return an error
		txResults := new(storage.TransactionResults)
		txResults.On("ByBlockIDTransactionID", bID, wrongTxID).Return(nil, status.Error(codes.Internal, "")).Once()

		handler := createHandler(txResults, true)

		_, err := handler.GetTransactionResult(context.Background(), req)

		// check that an error was received
		suite.Require().Error(err)
		errors.Is(err, status.Error(codes.Internal, ""))

		// check that one storage call was made
		suite.events.AssertExpectations(suite.T())
	})

	// failure path - method should fail if queries are disabled
	suite.Run("request with queries disabled", func() {

		req := createRequest(bID[:], txID[:])

		txResults := new(storage.TransactionResults)

		handler := createHandler(txResults, false)

		_, err := handler.GetTransactionResult(context.Background(), req)

		// check that an UNIMPLEMENTED error was received
		suite.Require().Error(err)
		suite.Assert().Equal(codes.Unimplemented, status.Code(err))
	})
}
